import os
import torch
import random
import numpy as np
from pathlib import Path
from PIL import Image
import cv2
from pyzjr.nn.torchutils.OneHot import get_one_hot
from torch.utils.data import Dataset
from pyzjr.data.__tensor import img2tensor, label2tensor

__all__ = ["BaseDataset", "BaseClsDataset", "SegmentTwoDataset", "VOCSegmentDataset"]

class BaseDataset(Dataset):
    """A simple dataset class with the same functionality as torch.utils.data.Dataset"""
    def __init__(self):
        pass

    def __getitem__(self, idx):
        pass

    def __len__(self):
        pass

    def list_files(self, imagefolder, image_extensions=('.jpg', '.jpeg', '.png', '.bmp')):
        """
        返回给定目录中的所有图像文件
        """
        return [
            f for f in Path(imagefolder).iterdir()
            if f.is_file() and not f.name.startswith(".") and f.suffix.lower() in image_extensions
        ]

    def list_dirs(self, imagefolder):
        """
        返回给定目录中的所有目录
        """
        return [f for f in Path(imagefolder).iterdir() if f.is_dir()]

    def rand(self, a=0., b=1.):
        """
        生成范围在 [a, b) 之间的随机数
        """
        return np.random.rand() * (b - a) + a

    def osmkdirs(self, *args):
        """
        如果不存在的话,创建多个目录
        """
        for path in args:
            if not os.path.exists(path):
                os.makedirs(path)

    def preprocess_input(self, image):
        """
        将图像转换为浮点数并进行归一化
        """
        image = np.array(image).astype(np.float32)
        image /= 255.0
        return image

    def check_input(self, input, classes=(float, int, int, str, list, tuple, np.ndarray, torch.Tensor)):
        """
        检查数据类型是否符合指定的类别
        classes: float, int, int, str, list, tuple, np.ndarray, torch.Tensor
        """
        return isinstance(input, classes)

    def Shuffle(self, files):
        """
        随机打乱文件列表
        """
        assert self.check_input(files, classes=(list, tuple)), "Input must be a list or tuple"
        random.shuffle(files)
        return files

    def load_txt_path_or_list(self, txt_path):
        """
        加载txt文件或列表,按行读取,去掉'/n'
        """
        path_list = []
        if self.check_input(txt_path, str):  # Check if txt_path is a string (path to a file)
            with open(txt_path, 'r') as file:
                lines = file.readlines()
                path_list.extend(line.strip() for line in lines)
        elif self.check_input(txt_path, list):  # Check if txt_path is a list
            path_list.extend(item.strip() for item in txt_path)
        else:
            raise ValueError(f"Invalid input type. {txt_path} should be a string (file path) or a list.")
        return path_list

    def imlabel2tensor(self, im, label, num_classes, sigmoid, totensor=False):
        im = img2tensor(im, totensor)
        label = label2tensor(label, num_classes, sigmoid, totensor)
        return im, label

class BaseClsDataset(BaseDataset):
    def __init__(self, file_txt, image_transform=None):
        """
        Supports txt file paths that match the format, or have been processed into the format list below.
        file_txt format: [path/to/xxx.jpg 0,
                          path/to/xxx.jpg 1,           OR           path/to/xxx.txt
                          path/to/xxx.jpg 2,
                          ...]
        """
        super().__init__()
        file_txt_result = self.load_txt_path_or_list(file_txt)
        if self._check_file_txt_format(file_txt_result):
            self.file_txt = file_txt_result
        self.image_transform = image_transform

    def __len__(self):
        return len(self.file_txt)

    def _check_file_txt_format(self, file_txt):
        for line in file_txt:
            parts = line.split()
            assert len(parts) == 2, f"BaseClsDataset: Invalid format in line: {line}"
            file_path, digit_label = parts
            assert os.path.exists(file_path), f"BaseClsDataset: File not found: {file_path}"
            assert digit_label.isdigit(), f"BaseClsDataset: Invalid digit label: {digit_label}"
        return True

    def __getitem__(self, idx):
        print(self.file_txt[idx])
        file_path, digit_label = self.file_txt[idx].split()
        raw_image = Image.open(file_path)
        image = raw_image.convert("RGB")
        if self.image_transform is not None:
            image = self.image_transform(image)
        return image, int(digit_label)

class VOCSegmentDataset(BaseDataset):
    def __init__(self,voc_root='VOCdevkit', year='2007', image_transform=None, txt_path="train.txt"):
        super(VOCSegmentDataset, self).__init__()
        assert year in ["2007","2012"], "year can only choose 2007 and 2012"
        root = os.path.join(voc_root, f"VOC{year}")
        assert os.path.exists(root), "path '{}' does not exist.".format(root)

        imgs_dir = os.path.join(root, 'JPEGImages')
        masks_dir = os.path.join(root, 'SegmentationClass')
        txt_path = os.path.join(root, "ImageSets", "Segmentation", txt_path)
        assert os.path.exists(txt_path), "file '{}' does not exist.".format(txt_path)
        with open(os.path.join(txt_path), "r") as f:
            file_names = [x.strip() for x in f.readlines() if len(x.strip()) > 0]

        self.images = [os.path.join(imgs_dir, x + ".jpg") for x in file_names]
        self.masks = [os.path.join(masks_dir, x + ".png") for x in file_names]
        assert (len(self.images) == len(self.masks))
        self.transforms = image_transform

    def __len__(self):
        return len(self.images)

    def __getitem__(self, index):
        """
        Args:
            index (int): Index
        Returns:
            tuple: (image, target) where target is the image segmentation.
        """
        img = Image.open(self.images[index]).convert("RGB")
        target = Image.open(self.masks[index]).convert("RGB")

        if self.transforms is not None:
            img, target = self.transforms(img, target)

        return img, target

class SegmentTwoDataset(BaseDataset):
    """
    二分类专用 Dataset ,目录结构请参照以下所示
    Example:
        pneumonia
            - image
            - label
            - test
            - train.txt
            - val.txt
            - test.txt
    """
    def __init__(self, file_path, input_shape=(512, 512), num_classes=2, train=True):
        super(SegmentTwoDataset, self).__init__()
        self.num_classes = num_classes
        self.input_shape = input_shape
        TXT = 'train.txt' if train else 'val.txt'
        txt_path = os.path.join(file_path, TXT)
        self.train = train

        self.image_paths = self.load_txt_path_or_list(txt_path)
        self.image_file_path = os.path.join(file_path, "image")
        self.label_file_path = os.path.join(file_path, "label")
        self.osmkdirs(file_path, self.label_file_path, self.image_file_path)

    def __len__(self):
        return len(self.image_paths)

    def __getitem__(self, index):
        one_image_name = self.image_paths[index]
        jpg = os.path.join(self.image_file_path, one_image_name) + '.jpg'
        png = os.path.join(self.label_file_path, one_image_name) + '.png'
        jpg, png = Image.open(jpg).convert("RGB"), Image.open(png).convert("L")
        jpg, png = self.get_random_data(jpg, png, self.input_shape, random=self.train)
        jpg, modify_png = self.imlabel2tensor(jpg, png, num_classes=2, sigmoid=False)
        seg_labels = modify_png
        # 在这里需要 + 1, 是因为数据集有些标签具有白边部分, 我们需要将白边部分进行忽略，+1的目的是方便忽略。
        seg_labels = get_one_hot(seg_labels, num_classes=self.num_classes + 1)
        return jpg, modify_png, seg_labels

    def get_image_name_list(self):
        return self.image_paths

    def get_random_data(self, image, label, input_shape, jitter=.3, hue=.1, sat=0.7, val=0.3, random=True):
        label = Image.fromarray(np.array(label))
        iw, ih = image.size
        h, w = input_shape

        if not random:
            iw, ih = image.size
            scale = min(w / iw, h / ih)
            nw = int(iw * scale)
            nh = int(ih * scale)

            image = image.resize((nw, nh), Image.BICUBIC)
            new_image = Image.new('RGB', (w, h), (128, 128, 128))
            new_image.paste(image, ((w - nw) // 2, (h - nh) // 2))

            label = label.resize((nw, nh), Image.NEAREST)
            new_label = Image.new('L', (w, h), (0))
            new_label.paste(label, ((w - nw) // 2, (h - nh) // 2))
            return new_image, new_label

        new_ar = iw / ih * self.rand(1 - jitter, 1 + jitter) / self.rand(1 - jitter, 1 + jitter)
        scale = self.rand(0.25, 2)
        if new_ar < 1:
            nh = int(scale * h)
            nw = int(nh * new_ar)
        else:
            nw = int(scale * w)
            nh = int(nw / new_ar)
        image = image.resize((nw, nh), Image.BICUBIC)
        label = label.resize((nw, nh), Image.NEAREST)

        flip = self.rand() < .5
        if flip:
            image = image.transpose(Image.FLIP_LEFT_RIGHT)
            label = label.transpose(Image.FLIP_LEFT_RIGHT)

        dx = int(self.rand(0, w - nw))
        dy = int(self.rand(0, h - nh))
        new_image = Image.new('RGB', (w, h), (128, 128, 128))
        new_label = Image.new('L', (w, h), (0))
        new_image.paste(image, (dx, dy))
        new_label.paste(label, (dx, dy))
        image = new_image
        label = new_label

        image_data = np.array(image, np.uint8)
        r = np.random.uniform(-1, 1, 3) * [hue, sat, val] + 1
        hue, sat, val = cv2.split(cv2.cvtColor(image_data, cv2.COLOR_RGB2HSV))
        dtype = image_data.dtype
        x = np.arange(0, 256, dtype=r.dtype)
        lut_hue = ((x * r[0]) % 180).astype(dtype)
        lut_sat = np.clip(x * r[1], 0, 255).astype(dtype)
        lut_val = np.clip(x * r[2], 0, 255).astype(dtype)
        image_data = cv2.merge((cv2.LUT(hue, lut_hue), cv2.LUT(sat, lut_sat), cv2.LUT(val, lut_val)))
        image_data = cv2.cvtColor(image_data, cv2.COLOR_HSV2RGB)

        return image_data, label

class YOLOv5Dataset(BaseDataset):
    def __init__(self,
                 path,
                 img_size=640,
                 batch_size=2,
                 image_transform=None,
                 ):
        super().__init__()
        self.path = path
        self.img_size = img_size
        self.batch_size = batch_size
        self.transform = image_transform
        self.img_files = self._check_path_is_list(self.path)
        self.label_files = self.img_to_label_paths(self.img_files)

    def img_to_label_paths(self, img_paths):
        """
        1 Replace '/images/' with '/labels/'.
        2 Remove the file extension by calling. split ('.', 1) [0].
        3 Finally, add. txt to obtain the final label path.
        """
        sa, sb = f'{os.sep}images{os.sep}', f'{os.sep}labels{os.sep}'  # /images/, /labels/ substrings
        return [sb.join(x.rsplit(sa, 1)).rsplit('.', 1)[0] + '.txt' for x in img_paths]

    def _check_path_is_list(self, _path):
        img_path_list = []
        if isinstance(_path, list):
            img_path_list.extend(item.strip() for item in _path)

if __name__ == "__main__":
    from torch.utils.data import DataLoader

    dataset_path = r"D:\PythonProject\uhrnet\data\pneumonia"
    train_dataset = SegmentTwoDataset(dataset_path, train=True)
    val_dataset = SegmentTwoDataset(dataset_path, train=False)

    train_loader = DataLoader(train_dataset, shuffle=True, batch_size=2, num_workers=8, pin_memory=True,
                              drop_last=True)
    val_loader = DataLoader(val_dataset, shuffle=True, batch_size=1, num_workers=4, pin_memory=True,
                            drop_last=True)

    for im, label, seglabel in train_loader:
        print(im.shape, label.shape, seglabel.shape)


from torch import nn
import torch.nn.init as init

def init_weights(model, conv=None, batchnorm=None, linear=None, lstm=None):
    """
    :param model: Pytorch Model which is nn.Module
    :param conv:  'kaiming' or 'xavier'
    :param batchnorm: 'normal' or 'constant'
    :param linear: 'kaiming' or 'xavier'
    :param lstm: 'kaiming' or 'xavier'
    """
    for m in model.modules():
        if conv is not None and isinstance(m, (nn.modules.conv._ConvNd)):
            if conv == 'kaiming':
                init.kaiming_normal_(m.weight)
            elif conv == 'xavier':
                init.xavier_normal_(m.weight)
            else:
                raise ValueError("init type of conv error.\n")
            if m.bias is not None:
                init.constant_(m.bias, 0)

        elif batchnorm is not None and isinstance(m, (nn.modules.batchnorm._BatchNorm)):
            if batchnorm == 'normal':
                init.normal_(m.weight, 1.0, 0.02)
            elif batchnorm == 'constant':
                init.constant_(m.weight, 1.0)
            else:
                raise ValueError("init type of batchnorm error.\n")
            init.constant_(m.bias, 0.0)

        elif linear is not None and isinstance(m, nn.Linear):
            if linear == 'kaiming':
                init.kaiming_normal_(m.weight)
            elif linear == 'xavier':
                init.xavier_normal_(m.weight)
            else:
                raise ValueError("init type of linear error.\n")
            if m.bias is not None:
                init.constant_(m.bias, 0)

        elif lstm is not None and isinstance(m, nn.LSTM):
            for name, param in m.named_parameters():
                if 'weight' in name:
                    if lstm == 'kaiming':
                        init.kaiming_normal_(param)
                    elif lstm == 'xavier':
                        init.xavier_normal_(param)
                    else:
                        raise ValueError("init type of lstm error.\n")
                elif 'bias' in name:
                    init.constant_(param, 0)

def official_init(model):
    # Official init from torch repo.
    for m in model.modules():
        if isinstance(m, nn.Conv2d):
            nn.init.kaiming_normal_(m.weight)
        elif isinstance(m, nn.BatchNorm2d):
            nn.init.constant_(m.weight, 1)
            nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.Linear):
            nn.init.constant_(m.bias, 0)


import torch
import random
import numpy as np
from tqdm import tqdm
from torch.utils.data import Dataset, DataLoader

__all__ = ["RepeatDataLoader", "Repeat_sampler", "seed_worker"]

def seed_worker(worker_id):
    # Set dataloader worker seed: https://pytorch.org/docs/stable/notes/randomness.html#dataloader
    worker_seed = torch.initial_seed() % 2 ** 32
    np.random.seed(worker_seed)
    random.seed(worker_seed)

class RepeatDataLoader(DataLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._DataLoader__initialized = False
        if self.batch_sampler is None:
            self.sampler = Repeat_sampler(self.sampler)
        else:
            self.batch_sampler = Repeat_sampler(self.batch_sampler)
        self._DataLoader__initialized = True
        self.iterator = super().__iter__()

    def __len__(self):
        return len(self.sampler) if self.batch_sampler is None else len(self.batch_sampler.sampler)

    def __iter__(self):
        total_iterations = len(self)
        for i in tqdm(range(total_iterations), desc=f"Iteration", unit="iteration"):
            yield next(self.iterator)

class Repeat_sampler(object):
    """ Function to create a sampler that repeats forever.

    Args:
        sampler (Sampler)
    """
    def __init__(self, sampler):
        self.sampler = sampler

    def __iter__(self):
        while True:
            yield from iter(self.sampler)





"""
time: 2024-02-19
处于大三的寒假, 正在看张宇的考研数学, 空余时间复现的泰勒展开式近似重要函数, 以及书上涉及的一些重要函数
参考书籍《张宇考研数学基础30讲》 P-27 (2025版)
"""
import math
import pyzjr.Math.constant as const
from pyzjr.Math.arithmetic import odd_factorial, even_factorial

__all__ = ["cos", "sin", "tan", "cot", "sec", "csc", "arcsin", "arccos", "arctan", "angle_to_2pi",
           "to_degree", "to_radians", "exp", "log_e", "sinh", "arsinh", "cosh", "arcosh",
           "sigmoid", "tanh", "artanh", "relu", "taylor_sin", "taylor_cos", "taylor_arcsin",
           "taylor_tan", "taylor_arctan", "taylor_log_1_plus_x", "taylor_exp",
           "taylor_series_1_plus_x_to_a"]

def cos(x):
    """弧度制 cos"""
    return math.cos(x)

def sin(x):
    """弧度制 sin"""
    return math.sin(x)

def tan(x):
    """弧度制 tan"""
    return math.tan(x)

def cot(x):
    tan_value = tan(x)
    if tan_value == 0:
        raise ValueError("cot(x) is undefined when tan(x) is zero.")
    return 1 / tan_value

def sec(x):
    """弧度制 sec"""
    cos_value = cos(x)
    if cos_value == 0:
        raise ValueError("sec(x) is undefined when cos(x) is zero.")
    return 1 / cos_value

def csc(x):
    """弧度制 csc"""
    sin_value = sin(x)
    if sin_value == 0:
        raise ValueError("csc(x) is undefined when sin(x) is zero.")
    return 1 / sin_value

def arccos(x):
    """计算反正弦值"""
    return math.acos(x)

def arcsin(x):
    """计算反正弦值"""
    return math.asin(x)

def arctan(x):
    """计算反正弦值"""
    return math.atan(x)

def angle_to_2pi(angle):
    """将角度映射到0到2π"""
    two_pi = 2 * const.pi
    return angle % two_pi

def to_degree(radians_value):
    """弧度转度数"""
    return math.degrees(radians_value)

def to_radians(angle_degrees):
    """度数转弧度"""
    return math.radians(angle_degrees)

def exp(x):
    """计算以e为底的指数函数值"""
    return (const.e) ** x

def log_e(x):
    """计算以e为底的对数函数值"""
    return math.log(x)

def sinh(x):
    """双曲正弦函数"""
    return (exp(x) - exp(-x)) / 2

def arsinh(x):
    """反双曲正弦函数"""
    return log_e(x + math.sqrt(x**2 + 1))

def cosh(x):
    """双曲余弦函数"""
    return (exp(x) + exp(-x)) / 2

def arcosh(x):
    """反双曲余弦函数"""
    return log_e(x + math.sqrt(x**2 - 1))

def tanh(x):
    """双曲正切函数"""
    return (exp(x) - exp(-x)) / (exp(x) + exp(-x))

def artanh(x):
    """反双曲正切函数"""
    return 0.5 * log_e((1+x) / (1-x))

def sigmoid(x):
    """Sigmoid 函数"""
    return 1 / (1 + exp(-x))

def relu(x):
    """ReLU 激活函数"""
    return max(0, x)

def taylor_exp(x, n):
    """泰勒近似 exp
    Example:
        approximation = taylor_exp(2, 20)
        exact_value = math.exp(2)
        print("Approximation:", approximation)
        print("Exact Value:", exact_value)
    """
    result = 0
    for i in range(n):
        result += (x ** i) / math.factorial(i)
    return result


def taylor_sin(x, n):
    """泰勒近似 sin
    Example:
        angle_degrees = 30
        angle_radians = math.radians(angle_degrees)
        approximation = taylor_sin(angle_radians, 5)
        exact_value = math.sin(angle_radians)
        print("Approximation:", approximation)
        print("Exact Value:", exact_value)
    """
    result = 0
    for i in range(n):
        term = ((-1) ** i) * (x ** (2*i + 1)) / math.factorial(2*i + 1)
        result += term
    return result


def taylor_cos(x, n):
    """泰勒近似 cos
    Example:
        angle_degrees = 60
        angle_radians = math.radians(angle_degrees)
        approximation = taylor_cos(angle_radians, 5)
        exact_value = math.cos(angle_radians)
        print("Approximation:", approximation)
        print("Exact Value:", exact_value)
    """
    result = 1
    for i in range(1, n):
        term = ((-1) ** i) * (x ** (2*i)) / math.factorial(2*i)
        result += term
    return result

def taylor_tan(x, n):
    """泰勒近似 tan
    因为分子分母无规律，这里只提供了前24项
    Example:
        angle_degrees = 45
        angle_radians = math.radians(angle_degrees)
        approximation = taylor_tan(angle_radians, 24)
        exact_value = math.tan(angle_radians)
        print("Approximation:", approximation)
        print("Exact Value:", exact_value)
    """
    # 从 1 - 23 奇数项系数
    coefficients = [1, 1/3, 2/15, 17/315, 62/2835, 1382/155925, 21844/6081075, 929569/638512875,
                    6404582/10854718875, 443861162/1856156927625, 18888466084/194896477400625,
                    113927491862/49308808782358125]
    result = 0
    iter = 0
    for i in range(n):
        if i <= 24 and i % 2 == 1:
            coefficient = coefficients[iter]
            iter += 1
            term = coefficient * (x ** i)
            result += term
    return result

def taylor_log_1_plus_x(x, n):
    """泰勒展开 In(1+x)
    Example:
        approximation = taylor_log_1_plus_x(0.5, 10)
        exact_value = math.log1p(0.5)
        print("Approximation:", approximation)
        print("Exact Value:", exact_value)
    """
    result = 0
    for i in range(1, n + 1):
        term = ((-1)**(i-1) * x**i) / i
        result += term
    return result

def taylor_arcsin(x, n):
    """泰勒展开 arcsinx
    x 属于 (-1, 1)
    Example:
        approximation = taylor_arcsin(0.5, 10)
        exact_value = math.asin(0.5)
        print("Approximation:", approximation)
        print("Exact Value:", exact_value)
    """
    if abs(x) > 1:
        raise ValueError("Input x must be between -1 and 1 for arcsin(x)")

    result = x
    for i in range(1, n+1):
        odd = 2 * i + 1
        term = ((x ** odd) / odd) * (odd_factorial(i) / even_factorial(i))
        result += term

    return result

def taylor_arctan(x, n):
    """泰勒展开 arctan(x)
    x 属于 (-1, 1)
    Example:
        approximation = taylor_arctan(0.5, 10)
        exact_value = math.atan(0.5)
        print("Approximation:", approximation)
        print("Exact Value:", exact_value)
    """
    if abs(x) > 1:
        raise ValueError("Input x must be between -1 and 1 for arctan(x)")

    result = 0
    for i in range(n):
        term = ((-1)**i * x**(2 * i + 1)) / (2 * i + 1)
        result += term

    return result

def taylor_series_1_plus_x_to_a(x, a, n):
    """(1+x)^a 的泰勒展开
    a 整数近似才比较相近, 用浮点数需要四舍五入
    Example:
        a = 3
        approximation = taylor_series_1_plus_x_to_a(0.5, a, 10)
        exact_value = (1 + 0.5) ** a
        print(f"Approximation: {approximation}, Exact Value: {exact_value}")
    """
    def binomial_coefficient(a, n):
        a = round(a)
        if n < 0 or n > a:
            return 0
        return math.factorial(a) // (math.factorial(n) * math.factorial(a - n))
    result = 0
    for i in range(n + 1):
        term = binomial_coefficient(a, i) * x**i
        result += term
    return result




import math
import numpy as np

npi = np.pi
two_npi = 2 * np.pi

def rand(a=0., b=1.):
    """
    生成在指定范围内的随机浮点数,进行缩放和偏移来映射到[a, b)的范围
    a: 下界; b: 上界
    """
    return np.random.rand() * (b - a) + a

def normal(x, mu, sigma):
    """正态分布（高斯分布）概率密度函数"""
    p = 1 / np.sqrt(two_npi * sigma**2)
    return p * np.exp(-0.5 / sigma**2 * (x - mu)**2)

def gaussian2D(x, y, sigma):
    # 计算二维高斯函数的数值
    normalization = 1 / ((two_npi**(0.5)) * sigma)
    exponent = -((x**2 + y**2) / (2 * sigma**2))
    result = normalization * np.exp(exponent)
    return result